package com.sun.org.apache.xerces.internal.impl.xs;

import java.util.List;
import java.util.Collections;
import java.util.ArrayList;
import com.sun.org.apache.xerces.internal.impl.xs.models.XSCMValidator;
import com.sun.org.apache.xerces.internal.util.SymbolHash;
import com.sun.org.apache.xerces.internal.impl.xs.models.CMBuilder;
import com.sun.org.apache.xerces.internal.impl.xs.util.SimpleLocator;
import com.sun.org.apache.xerces.internal.impl.XMLErrorReporter;
import com.sun.org.apache.xerces.internal.impl.dv.InvalidDatatypeValueException;
import com.sun.org.apache.xerces.internal.impl.dv.ValidatedInfo;
import com.sun.org.apache.xerces.internal.impl.dv.ValidationContext;
import com.sun.org.apache.xerces.internal.xs.XSObjectList;
import com.sun.org.apache.xerces.internal.xs.XSTypeDefinition;
import com.sun.org.apache.xerces.internal.impl.xs.util.XSObjectListImpl;
import java.util.Comparator;
import com.sun.org.apache.xerces.internal.impl.dv.XSSimpleType;

public class XSConstraints
{
    static final int OCCURRENCE_UNKNOWN = -2;
    static final XSSimpleType STRING_TYPE;
    private static XSParticleDecl fEmptyParticle;
    private static final Comparator<XSParticleDecl> ELEMENT_PARTICLE_COMPARATOR;
    
    public XSConstraints() {
    }
    
    public static XSParticleDecl getEmptySequence() {
        if (XSConstraints.fEmptyParticle == null) {
            XSModelGroupImpl group = new XSModelGroupImpl();
            group.fCompositor = 102;
            group.fParticleCount = 0;
            group.fParticles = null;
            group.fAnnotations = XSObjectListImpl.EMPTY_LIST;
            XSParticleDecl particle = new XSParticleDecl();
            particle.fType = 3;
            particle.fValue = group;
            particle.fAnnotations = XSObjectListImpl.EMPTY_LIST;
            XSConstraints.fEmptyParticle = particle;
        }
        return XSConstraints.fEmptyParticle;
    }
    
    public static boolean checkTypeDerivationOk(XSTypeDefinition derived, XSTypeDefinition base, short block) {
        if (derived == SchemaGrammar.fAnyType) {
            return derived == base;
        }
        if (derived == SchemaGrammar.fAnySimpleType) {
            return base == SchemaGrammar.fAnyType || base == SchemaGrammar.fAnySimpleType;
        }
        if (derived.getTypeCategory() == 16) {
            if (base.getTypeCategory() == 15) {
                if (base == SchemaGrammar.fAnyType) {
                    base = SchemaGrammar.fAnySimpleType;
                }
                else {
                    return false;
                }
            }
            return checkSimpleDerivation((XSSimpleType)derived, (XSSimpleType)base, block);
        }
        else {
            return checkComplexDerivation((XSComplexTypeDecl)derived, base, block);
        }
    }
    
    public static boolean checkSimpleDerivationOk(XSSimpleType derived, XSTypeDefinition base, short block) {
        if (derived == SchemaGrammar.fAnySimpleType) {
            return base == SchemaGrammar.fAnyType || base == SchemaGrammar.fAnySimpleType;
        }
        if (base.getTypeCategory() == 15) {
            if (base == SchemaGrammar.fAnyType) {
                base = SchemaGrammar.fAnySimpleType;
            }
            else {
                return false;
            }
        }
        return checkSimpleDerivation(derived, (XSSimpleType)base, block);
    }
    
    public static boolean checkComplexDerivationOk(XSComplexTypeDecl derived, XSTypeDefinition base, short block) {
        if (derived == SchemaGrammar.fAnyType) {
            return derived == base;
        }
        return checkComplexDerivation(derived, base, block);
    }
    
    private static boolean checkSimpleDerivation(XSSimpleType derived, XSSimpleType base, short block) {
        if (derived == base) {
            return true;
        }
        if ((block & 0x2) != 0x0 || (derived.getBaseType().getFinal() & 0x2) != 0x0) {
            return false;
        }
        XSSimpleType directBase = (XSSimpleType)derived.getBaseType();
        if (directBase == base) {
            return true;
        }
        if (directBase != SchemaGrammar.fAnySimpleType && checkSimpleDerivation(directBase, base, block)) {
            return true;
        }
        if ((derived.getVariety() == 2 || derived.getVariety() == 3) && base == SchemaGrammar.fAnySimpleType) {
            return true;
        }
        if (base.getVariety() == 3) {
            XSObjectList subUnionMemberDV = base.getMemberTypes();
            int subUnionSize = subUnionMemberDV.getLength();
            int i = 0;
            while (i < subUnionSize) {
                base = (XSSimpleType)subUnionMemberDV.item(i);
                if (checkSimpleDerivation(derived, base, block)) {
                    return true;
                }
                ++i;
            }
        }
        return false;
    }
    
    private static boolean checkComplexDerivation(XSComplexTypeDecl derived, XSTypeDefinition base, short block) {
        if (derived == base) {
            return true;
        }
        if ((derived.fDerivedBy & block) != 0x0) {
            return false;
        }
        XSTypeDefinition directBase = derived.fBaseType;
        if (directBase == base) {
            return true;
        }
        if (directBase == SchemaGrammar.fAnyType || directBase == SchemaGrammar.fAnySimpleType) {
            return false;
        }
        if (directBase.getTypeCategory() == 15) {
            return checkComplexDerivation((XSComplexTypeDecl)directBase, base, block);
        }
        if (directBase.getTypeCategory() == 16) {
            if (base.getTypeCategory() == 15) {
                if (base == SchemaGrammar.fAnyType) {
                    base = SchemaGrammar.fAnySimpleType;
                }
                else {
                    return false;
                }
            }
            return checkSimpleDerivation((XSSimpleType)directBase, (XSSimpleType)base, block);
        }
        else {
            return false;
        }
    }
    
    public static Object ElementDefaultValidImmediate(XSTypeDefinition type, String value, ValidationContext context, ValidatedInfo vinfo) {
        XSSimpleType dv = null;
        if (type.getTypeCategory() == 16) {
            dv = (XSSimpleType)type;
        }
        else {
            XSComplexTypeDecl ctype = (XSComplexTypeDecl)type;
            if (ctype.fContentType == 1) {
                dv = ctype.fXSSimpleType;
            }
            else if (ctype.fContentType == 3) {
                if (!((XSParticleDecl)ctype.getParticle()).emptiable()) {
                    return null;
                }
            }
            else {
                return null;
            }
        }
        Object actualValue = null;
        if (dv == null) {
            dv = XSConstraints.STRING_TYPE;
        }
        try {
            actualValue = dv.validate(value, context, vinfo);
            if (vinfo != null) {
                actualValue = dv.validate(vinfo.stringValue(), context, vinfo);
            }
        }
        catch (InvalidDatatypeValueException ide) {
            return null;
        }
        return actualValue;
    }
    
    static void reportSchemaError(XMLErrorReporter errorReporter, SimpleLocator loc, String key, Object[] args) {
        if (loc != null) {
            errorReporter.reportError(loc, "http://www.w3.org/TR/xml-schema-1", key, args, (short)1);
        }
        else {
            errorReporter.reportError("http://www.w3.org/TR/xml-schema-1", key, args, (short)1);
        }
    }
    
    public static void fullSchemaChecking(XSGrammarBucket grammarBucket, SubstitutionGroupHandler SGHandler, CMBuilder cmBuilder, XMLErrorReporter errorReporter) {
        SchemaGrammar[] grammars = grammarBucket.getGrammars();
        for (int i = grammars.length - 1; i >= 0; --i) {
            SGHandler.addSubstitutionGroup(grammars[i].getSubstitutionGroups());
        }
        XSParticleDecl fakeDerived = new XSParticleDecl();
        XSParticleDecl fakeBase = new XSParticleDecl();
        fakeDerived.fType = 3;
        fakeBase.fType = 3;
        for (int g = grammars.length - 1; g >= 0; --g) {
            XSGroupDecl[] redefinedGroups = grammars[g].getRedefinedGroupDecls();
            SimpleLocator[] rgLocators = grammars[g].getRGLocators();
            int j = 0;
            while (j < redefinedGroups.length) {
                XSGroupDecl derivedGrp = redefinedGroups[j++];
                XSModelGroupImpl derivedMG = derivedGrp.fModelGroup;
                XSGroupDecl baseGrp = redefinedGroups[j++];
                XSModelGroupImpl baseMG = baseGrp.fModelGroup;
                fakeDerived.fValue = derivedMG;
                if ((fakeBase.fValue = baseMG) == null) {
                    if (derivedMG != null) {
                        reportSchemaError(errorReporter, rgLocators[j / 2 - 1], "src-redefine.6.2.2", new Object[] { derivedGrp.fName, "rcase-Recurse.2" });
                    }
                    else {
                        continue;
                    }
                }
                else if (derivedMG == null) {
                    if (!fakeBase.emptiable()) {
                        reportSchemaError(errorReporter, rgLocators[j / 2 - 1], "src-redefine.6.2.2", new Object[] { derivedGrp.fName, "rcase-Recurse.2" });
                    }
                    else {
                        continue;
                    }
                }
                else {
                    try {
                        particleValidRestriction(fakeDerived, SGHandler, fakeBase, SGHandler);
                    }
                    catch (XMLSchemaException e) {
                        String key = e.getKey();
                        reportSchemaError(errorReporter, rgLocators[j / 2 - 1], key, e.getArgs());
                        reportSchemaError(errorReporter, rgLocators[j / 2 - 1], "src-redefine.6.2.2", new Object[] { derivedGrp.fName, key });
                    }
                }
            }
        }
        SymbolHash elemTable = new SymbolHash();
        for (int k = grammars.length - 1; k >= 0; --k) {
            int keepType = 0;
            boolean fullChecked = grammars[k].fFullChecked;
            XSComplexTypeDecl[] types = grammars[k].getUncheckedComplexTypeDecls();
            SimpleLocator[] ctLocators = grammars[k].getUncheckedCTLocators();
            for (int l = 0; l < types.length; ++l) {
                if (!fullChecked && types[l].fParticle != null) {
                    elemTable.clear();
                    try {
                        checkElementDeclsConsistent(types[l], types[l].fParticle, elemTable, SGHandler);
                    }
                    catch (XMLSchemaException e) {
                        reportSchemaError(errorReporter, ctLocators[l], e.getKey(), e.getArgs());
                    }
                }
                if (types[l].fBaseType != null && types[l].fBaseType != SchemaGrammar.fAnyType && types[l].fDerivedBy == 2 && types[l].fBaseType instanceof XSComplexTypeDecl) {
                    XSParticleDecl derivedParticle = types[l].fParticle;
                    XSParticleDecl baseParticle = ((XSComplexTypeDecl)(XSComplexTypeDecl)types[l].fBaseType).fParticle;
                    if (derivedParticle == null) {
                        if (baseParticle != null && !baseParticle.emptiable()) {
                            reportSchemaError(errorReporter, ctLocators[l], "derivation-ok-restriction.5.3.2", new Object[] { types[l].fName, types[l].fBaseType.getName() });
                        }
                    }
                    else if (baseParticle != null) {
                        try {
                            particleValidRestriction(types[l].fParticle, SGHandler, ((XSComplexTypeDecl)(XSComplexTypeDecl)types[l].fBaseType).fParticle, SGHandler);
                        }
                        catch (XMLSchemaException e2) {
                            reportSchemaError(errorReporter, ctLocators[l], e2.getKey(), e2.getArgs());
                            reportSchemaError(errorReporter, ctLocators[l], "derivation-ok-restriction.5.4.2", new Object[] { types[l].fName });
                        }
                    }
                    else {
                        reportSchemaError(errorReporter, ctLocators[l], "derivation-ok-restriction.5.4.2", new Object[] { types[l].fName });
                    }
                }
                XSCMValidator cm = types[l].getContentModel(cmBuilder, true);
                boolean further = false;
                if (cm != null) {
                    try {
                        further = cm.checkUniqueParticleAttribution(SGHandler);
                    }
                    catch (XMLSchemaException e3) {
                        reportSchemaError(errorReporter, ctLocators[l], e3.getKey(), e3.getArgs());
                    }
                }
                if (!fullChecked && further) {
                    types[keepType++] = types[l];
                }
            }
            if (!fullChecked) {
                grammars[k].setUncheckedTypeNum(keepType);
                grammars[k].fFullChecked = true;
            }
        }
    }
    
    public static void checkElementDeclsConsistent(XSComplexTypeDecl type, XSParticleDecl particle, SymbolHash elemDeclHash, SubstitutionGroupHandler sgHandler) throws XMLSchemaException {
        int pType = particle.fType;
        if (pType == 2) {
            return;
        }
        if (pType == 1) {
            XSElementDecl elem = (XSElementDecl)(XSElementDecl)particle.fValue;
            findElemInTable(type, elem, elemDeclHash);
            if (elem.fScope == 1) {
                XSElementDecl[] subGroup = sgHandler.getSubstitutionGroup(elem);
                for (int j = 0; j < subGroup.length; ++j) {
                    findElemInTable(type, subGroup[j], elemDeclHash);
                }
            }
            return;
        }
        else {
            XSModelGroupImpl group = (XSModelGroupImpl)particle.fValue;
            for (int i = 0; i < group.fParticleCount; ++i) {
                checkElementDeclsConsistent(type, group.fParticles[i], elemDeclHash, sgHandler);
            }
            return;
        }
    }
    
    public static void findElemInTable(XSComplexTypeDecl type, XSElementDecl elem, SymbolHash elemDeclHash) throws XMLSchemaException {
        String name = elem.fName + "," + elem.fTargetNamespace;
        XSElementDecl existingElem = null;
        if ((existingElem = (XSElementDecl)(XSElementDecl)elemDeclHash.get(name)) == null) {
            elemDeclHash.put(name, elem);
        }
        else {
            if (elem == existingElem) {
                return;
            }
            if (elem.fType != existingElem.fType) {
                throw new XMLSchemaException("cos-element-consistent", new Object[] { type.fName, elem.fName });
            }
        }
    }
    
    private static boolean particleValidRestriction(XSParticleDecl dParticle, SubstitutionGroupHandler dSGHandler, XSParticleDecl bParticle, SubstitutionGroupHandler bSGHandler) throws XMLSchemaException {
        return particleValidRestriction(dParticle, dSGHandler, bParticle, bSGHandler, true);
    }
    
    private static boolean particleValidRestriction(XSParticleDecl dParticle, SubstitutionGroupHandler dSGHandler, XSParticleDecl bParticle, SubstitutionGroupHandler bSGHandler, boolean checkWCOccurrence) throws XMLSchemaException {
        List<XSParticleDecl> dChildren = null;
        List<XSParticleDecl> bChildren = null;
        int dMinEffectiveTotalRange = -2;
        int dMaxEffectiveTotalRange = -2;
        boolean bExpansionHappened = false;
        if (dParticle.isEmpty() && !bParticle.emptiable()) {
            throw new XMLSchemaException("cos-particle-restrict.a", null);
        }
        if (!dParticle.isEmpty() && bParticle.isEmpty()) {
            throw new XMLSchemaException("cos-particle-restrict.b", null);
        }
        short dType = dParticle.fType;
        if (dType == 3) {
            dType = ((XSModelGroupImpl)dParticle.fValue).fCompositor;
            XSParticleDecl dtmp = getNonUnaryGroup(dParticle);
            if (dtmp != dParticle) {
                dParticle = dtmp;
                dType = dParticle.fType;
                if (dType == 3) {
                    dType = ((XSModelGroupImpl)dParticle.fValue).fCompositor;
                }
            }
            dChildren = removePointlessChildren(dParticle);
        }
        int dMinOccurs = dParticle.fMinOccurs;
        int dMaxOccurs = dParticle.fMaxOccurs;
        if (dSGHandler != null && dType == 1) {
            XSElementDecl dElement = (XSElementDecl)dParticle.fValue;
            if (dElement.fScope == 1) {
                XSElementDecl[] subGroup = dSGHandler.getSubstitutionGroup(dElement);
                if (subGroup.length > 0) {
                    dType = 101;
                    dMinEffectiveTotalRange = dMinOccurs;
                    dMaxEffectiveTotalRange = dMaxOccurs;
                    dChildren = new ArrayList(subGroup.length + 1);
                    for (int i = 0; i < subGroup.length; ++i) {
                        addElementToParticleVector(dChildren, subGroup[i]);
                    }
                    addElementToParticleVector(dChildren, dElement);
                    Collections.sort(dChildren, XSConstraints.ELEMENT_PARTICLE_COMPARATOR);
                    dSGHandler = null;
                }
            }
        }
        short bType = bParticle.fType;
        if (bType == 3) {
            bType = ((XSModelGroupImpl)bParticle.fValue).fCompositor;
            XSParticleDecl btmp = getNonUnaryGroup(bParticle);
            if (btmp != bParticle) {
                bParticle = btmp;
                bType = bParticle.fType;
                if (bType == 3) {
                    bType = ((XSModelGroupImpl)bParticle.fValue).fCompositor;
                }
            }
            bChildren = removePointlessChildren(bParticle);
        }
        int bMinOccurs = bParticle.fMinOccurs;
        int bMaxOccurs = bParticle.fMaxOccurs;
        if (bSGHandler != null && bType == 1) {
            XSElementDecl bElement = (XSElementDecl)bParticle.fValue;
            if (bElement.fScope == 1) {
                XSElementDecl[] bsubGroup = bSGHandler.getSubstitutionGroup(bElement);
                if (bsubGroup.length > 0) {
                    bType = 101;
                    bChildren = new ArrayList(bsubGroup.length + 1);
                    for (int j = 0; j < bsubGroup.length; ++j) {
                        addElementToParticleVector(bChildren, bsubGroup[j]);
                    }
                    addElementToParticleVector(bChildren, bElement);
                    Collections.sort(bChildren, XSConstraints.ELEMENT_PARTICLE_COMPARATOR);
                    bSGHandler = null;
                    bExpansionHappened = true;
                }
            }
        }
        switch (dType) {
            case 1: {
                switch (bType) {
                    case 1: {
                        checkNameAndTypeOK((XSElementDecl)dParticle.fValue, dMinOccurs, dMaxOccurs, (XSElementDecl)bParticle.fValue, bMinOccurs, bMaxOccurs);
                        return bExpansionHappened;
                    }
                    case 2: {
                        checkNSCompat((XSElementDecl)dParticle.fValue, dMinOccurs, dMaxOccurs, (XSWildcardDecl)bParticle.fValue, bMinOccurs, bMaxOccurs, checkWCOccurrence);
                        return bExpansionHappened;
                    }
                    case 101: {
                        dChildren = new ArrayList();
                        dChildren.add(dParticle);
                        checkRecurseLax(dChildren, 1, 1, dSGHandler, bChildren, bMinOccurs, bMaxOccurs, bSGHandler);
                        return bExpansionHappened;
                    }
                    case 102:
                    case 103: {
                        dChildren = new ArrayList();
                        dChildren.add(dParticle);
                        checkRecurse(dChildren, 1, 1, dSGHandler, bChildren, bMinOccurs, bMaxOccurs, bSGHandler);
                        return bExpansionHappened;
                    }
                    default: {
                        throw new XMLSchemaException("Internal-Error", new Object[] { "in particleValidRestriction" });
                    }
                }
            }
            case 2: {
                switch (bType) {
                    case 2: {
                        checkNSSubset((XSWildcardDecl)dParticle.fValue, dMinOccurs, dMaxOccurs, (XSWildcardDecl)bParticle.fValue, bMinOccurs, bMaxOccurs);
                        return bExpansionHappened;
                    }
                    case 1:
                    case 101:
                    case 102:
                    case 103: {
                        throw new XMLSchemaException("cos-particle-restrict.2", new Object[] { "any:choice,sequence,all,elt" });
                    }
                    default: {
                        throw new XMLSchemaException("Internal-Error", new Object[] { "in particleValidRestriction" });
                    }
                }
            }
            case 103: {
                switch (bType) {
                    case 2: {
                        if (dMinEffectiveTotalRange == -2) {
                            dMinEffectiveTotalRange = dParticle.minEffectiveTotalRange();
                        }
                        if (dMaxEffectiveTotalRange == -2) {
                            dMaxEffectiveTotalRange = dParticle.maxEffectiveTotalRange();
                        }
                        checkNSRecurseCheckCardinality(dChildren, dMinEffectiveTotalRange, dMaxEffectiveTotalRange, dSGHandler, bParticle, bMinOccurs, bMaxOccurs, checkWCOccurrence);
                        return bExpansionHappened;
                    }
                    case 103: {
                        checkRecurse(dChildren, dMinOccurs, dMaxOccurs, dSGHandler, bChildren, bMinOccurs, bMaxOccurs, bSGHandler);
                        return bExpansionHappened;
                    }
                    case 1:
                    case 101:
                    case 102: {
                        throw new XMLSchemaException("cos-particle-restrict.2", new Object[] { "all:choice,sequence,elt" });
                    }
                    default: {
                        throw new XMLSchemaException("Internal-Error", new Object[] { "in particleValidRestriction" });
                    }
                }
            }
            case 101: {
                switch (bType) {
                    case 2: {
                        if (dMinEffectiveTotalRange == -2) {
                            dMinEffectiveTotalRange = dParticle.minEffectiveTotalRange();
                        }
                        if (dMaxEffectiveTotalRange == -2) {
                            dMaxEffectiveTotalRange = dParticle.maxEffectiveTotalRange();
                        }
                        checkNSRecurseCheckCardinality(dChildren, dMinEffectiveTotalRange, dMaxEffectiveTotalRange, dSGHandler, bParticle, bMinOccurs, bMaxOccurs, checkWCOccurrence);
                        return bExpansionHappened;
                    }
                    case 101: {
                        checkRecurseLax(dChildren, dMinOccurs, dMaxOccurs, dSGHandler, bChildren, bMinOccurs, bMaxOccurs, bSGHandler);
                        return bExpansionHappened;
                    }
                    case 1:
                    case 102:
                    case 103: {
                        throw new XMLSchemaException("cos-particle-restrict.2", new Object[] { "choice:all,sequence,elt" });
                    }
                    default: {
                        throw new XMLSchemaException("Internal-Error", new Object[] { "in particleValidRestriction" });
                    }
                }
            }
            case 102: {
                switch (bType) {
                    case 2: {
                        if (dMinEffectiveTotalRange == -2) {
                            dMinEffectiveTotalRange = dParticle.minEffectiveTotalRange();
                        }
                        if (dMaxEffectiveTotalRange == -2) {
                            dMaxEffectiveTotalRange = dParticle.maxEffectiveTotalRange();
                        }
                        checkNSRecurseCheckCardinality(dChildren, dMinEffectiveTotalRange, dMaxEffectiveTotalRange, dSGHandler, bParticle, bMinOccurs, bMaxOccurs, checkWCOccurrence);
                        return bExpansionHappened;
                    }
                    case 103: {
                        checkRecurseUnordered(dChildren, dMinOccurs, dMaxOccurs, dSGHandler, bChildren, bMinOccurs, bMaxOccurs, bSGHandler);
                        return bExpansionHappened;
                    }
                    case 102: {
                        checkRecurse(dChildren, dMinOccurs, dMaxOccurs, dSGHandler, bChildren, bMinOccurs, bMaxOccurs, bSGHandler);
                        return bExpansionHappened;
                    }
                    case 101: {
                        int min1 = dMinOccurs * dChildren.size();
                        int max1 = (dMaxOccurs == -1) ? dMaxOccurs : (dMaxOccurs * dChildren.size());
                        checkMapAndSum(dChildren, min1, max1, dSGHandler, bChildren, bMinOccurs, bMaxOccurs, bSGHandler);
                        return bExpansionHappened;
                    }
                    case 1: {
                        throw new XMLSchemaException("cos-particle-restrict.2", new Object[] { "seq:elt" });
                    }
                    default: {
                        throw new XMLSchemaException("Internal-Error", new Object[] { "in particleValidRestriction" });
                    }
                }
            }
            default: {
                return bExpansionHappened;
            }
        }
    }
    
    private static void addElementToParticleVector(List<XSParticleDecl> v, XSElementDecl d) {
        XSParticleDecl p = new XSParticleDecl();
        p.fValue = d;
        p.fType = 1;
        v.add(p);
    }
    
    private static XSParticleDecl getNonUnaryGroup(XSParticleDecl p) {
        if (p.fType == 1 || p.fType == 2) {
            return p;
        }
        if (p.fMinOccurs == 1 && p.fMaxOccurs == 1 && p.fValue != null && ((XSModelGroupImpl)p.fValue).fParticleCount == 1) {
            return getNonUnaryGroup(((XSModelGroupImpl)p.fValue).fParticles[0]);
        }
        return p;
    }
    
    private static List<XSParticleDecl> removePointlessChildren(XSParticleDecl p) {
        if (p.fType == 1 || p.fType == 2) {
            return null;
        }
        List<XSParticleDecl> children = new ArrayList();
        XSModelGroupImpl group = (XSModelGroupImpl)p.fValue;
        for (int i = 0; i < group.fParticleCount; ++i) {
            gatherChildren((int)group.fCompositor, group.fParticles[i], children);
        }
        return children;
    }
    
    private static void gatherChildren(int parentType, XSParticleDecl p, List<XSParticleDecl> children) {
        int min = p.fMinOccurs;
        int max = p.fMaxOccurs;
        int type = p.fType;
        if (type == 3) {
            type = ((XSModelGroupImpl)p.fValue).fCompositor;
        }
        if (type == 1 || type == 2) {
            children.add(p);
            return;
        }
        if (min != 1 || max != 1) {
            children.add(p);
        }
        else if (parentType == type) {
            XSModelGroupImpl group = (XSModelGroupImpl)p.fValue;
            for (int i = 0; i < group.fParticleCount; ++i) {
                gatherChildren(type, group.fParticles[i], children);
            }
        }
        else if (!p.isEmpty()) {
            children.add(p);
        }
        return;
    }
    
    private static void checkNameAndTypeOK(XSElementDecl dElement, int dMin, int dMax, XSElementDecl bElement, int bMin, int bMax) throws XMLSchemaException {
        if (dElement.fName != bElement.fName || dElement.fTargetNamespace != bElement.fTargetNamespace) {
            throw new XMLSchemaException("rcase-NameAndTypeOK.1", new Object[] { dElement.fName, dElement.fTargetNamespace, bElement.fName, bElement.fTargetNamespace });
        }
        if (!bElement.getNillable() && dElement.getNillable()) {
            throw new XMLSchemaException("rcase-NameAndTypeOK.2", new Object[] { dElement.fName });
        }
        if (!checkOccurrenceRange(dMin, dMax, bMin, bMax)) {
            throw new XMLSchemaException("rcase-NameAndTypeOK.3", new Object[] { dElement.fName, Integer.toString(dMin), (dMax == -1) ? "unbounded" : Integer.toString(dMax), Integer.toString(bMin), (bMax == -1) ? "unbounded" : Integer.toString(bMax) });
        }
        if (bElement.getConstraintType() == 2) {
            if (dElement.getConstraintType() != 2) {
                throw new XMLSchemaException("rcase-NameAndTypeOK.4.a", new Object[] { dElement.fName, bElement.fDefault.stringValue() });
            }
            boolean isSimple = false;
            if (dElement.fType.getTypeCategory() == 16 || ((XSComplexTypeDecl)dElement.fType).fContentType == 1) {
                isSimple = true;
            }
            if ((!isSimple && !bElement.fDefault.normalizedValue.equals(dElement.fDefault.normalizedValue)) || (isSimple && !bElement.fDefault.actualValue.equals(dElement.fDefault.actualValue))) {
                throw new XMLSchemaException("rcase-NameAndTypeOK.4.b", new Object[] { dElement.fName, dElement.fDefault.stringValue(), bElement.fDefault.stringValue() });
            }
        }
        checkIDConstraintRestriction(dElement, bElement);
        int blockSet1 = dElement.fBlock;
        int blockSet2 = bElement.fBlock;
        if ((blockSet1 & blockSet2) != blockSet2 || (blockSet1 == 0 && blockSet2 != 0)) {
            throw new XMLSchemaException("rcase-NameAndTypeOK.6", new Object[] { dElement.fName });
        }
        if (!checkTypeDerivationOk(dElement.fType, bElement.fType, (short)25)) {
            throw new XMLSchemaException("rcase-NameAndTypeOK.7", new Object[] { dElement.fName, dElement.fType.getName(), bElement.fType.getName() });
        }
        return;
    }
    
    private static void checkIDConstraintRestriction(XSElementDecl derivedElemDecl, XSElementDecl baseElemDecl) throws XMLSchemaException {
    }
    
    private static boolean checkOccurrenceRange(int min1, int max1, int min2, int max2) {
        return min1 >= min2 && (max2 == -1 || (max1 != -1 && max1 <= max2));
    }
    
    private static void checkNSCompat(XSElementDecl elem, int min1, int max1, XSWildcardDecl wildcard, int min2, int max2, boolean checkWCOccurrence) throws XMLSchemaException {
        if (checkWCOccurrence && !checkOccurrenceRange(min1, max1, min2, max2)) {
            throw new XMLSchemaException("rcase-NSCompat.2", new Object[] { elem.fName, Integer.toString(min1), (max1 == -1) ? "unbounded" : Integer.toString(max1), Integer.toString(min2), (max2 == -1) ? "unbounded" : Integer.toString(max2) });
        }
        if (!wildcard.allowNamespace(elem.fTargetNamespace)) {
            throw new XMLSchemaException("rcase-NSCompat.1", new Object[] { elem.fName, elem.fTargetNamespace });
        }
        return;
    }
    
    private static void checkNSSubset(XSWildcardDecl dWildcard, int min1, int max1, XSWildcardDecl bWildcard, int min2, int max2) throws XMLSchemaException {
        if (!checkOccurrenceRange(min1, max1, min2, max2)) {
            throw new XMLSchemaException("rcase-NSSubset.2", new Object[] { Integer.toString(min1), (max1 == -1) ? "unbounded" : Integer.toString(max1), Integer.toString(min2), (max2 == -1) ? "unbounded" : Integer.toString(max2) });
        }
        if (!dWildcard.isSubsetOf(bWildcard)) {
            throw new XMLSchemaException("rcase-NSSubset.1", null);
        }
        if (dWildcard.weakerProcessContents(bWildcard)) {
            throw new XMLSchemaException("rcase-NSSubset.3", new Object[] { dWildcard.getProcessContentsAsString(), bWildcard.getProcessContentsAsString() });
        }
        return;
    }
    
    private static void checkNSRecurseCheckCardinality(List<XSParticleDecl> children, int min1, int max1, SubstitutionGroupHandler dSGHandler, XSParticleDecl wildcard, int min2, int max2, boolean checkWCOccurrence) throws XMLSchemaException {
        if (checkWCOccurrence && !checkOccurrenceRange(min1, max1, min2, max2)) {
            throw new XMLSchemaException("rcase-NSRecurseCheckCardinality.2", new Object[] { Integer.toString(min1), (max1 == -1) ? "unbounded" : Integer.toString(max1), Integer.toString(min2), (max2 == -1) ? "unbounded" : Integer.toString(max2) });
        }
        int count = children.size();
        try {
            for (int i = 0; i < count; ++i) {
                XSParticleDecl particle1 = (XSParticleDecl)children.get(i);
                particleValidRestriction(particle1, dSGHandler, wildcard, null, false);
            }
        }
        catch (XMLSchemaException e) {
            throw new XMLSchemaException("rcase-NSRecurseCheckCardinality.1", null);
        }
        return;
    }
    
    private static void checkRecurse(List<XSParticleDecl> dChildren, int min1, int max1, SubstitutionGroupHandler dSGHandler, List<XSParticleDecl> bChildren, int min2, int max2, SubstitutionGroupHandler bSGHandler) throws XMLSchemaException {
        if (!checkOccurrenceRange(min1, max1, min2, max2)) {
            throw new XMLSchemaException("rcase-Recurse.1", new Object[] { Integer.toString(min1), (max1 == -1) ? "unbounded" : Integer.toString(max1), Integer.toString(min2), (max2 == -1) ? "unbounded" : Integer.toString(max2) });
        }
        int count1 = dChildren.size();
        int count2 = bChildren.size();
        int current = 0;
        int i = 0;
    Label_0100:
        while (i < count1) {
            XSParticleDecl particle1 = (XSParticleDecl)dChildren.get(i);
            int j = current;
            while (j < count2) {
                XSParticleDecl particle2 = (XSParticleDecl)bChildren.get(j);
                ++current;
                try {
                    particleValidRestriction(particle1, dSGHandler, particle2, bSGHandler);
                }
                catch (XMLSchemaException e) {
                    if (!particle2.emptiable()) {
                        throw new XMLSchemaException("rcase-Recurse.2", null);
                    }
                    ++j;
                    continue;
                }
                ++i;
                continue Label_0100;
            }
            throw new XMLSchemaException("rcase-Recurse.2", null);
        }
        int k = current;
        while (k < count2) {
            XSParticleDecl particle3 = (XSParticleDecl)bChildren.get(k);
            if (!particle3.emptiable()) {
                throw new XMLSchemaException("rcase-Recurse.2", null);
            }
            ++k;
        }
        return;
    }
    
    private static void checkRecurseUnordered(List<XSParticleDecl> dChildren, int min1, int max1, SubstitutionGroupHandler dSGHandler, List<XSParticleDecl> bChildren, int min2, int max2, SubstitutionGroupHandler bSGHandler) throws XMLSchemaException {
        if (!checkOccurrenceRange(min1, max1, min2, max2)) {
            throw new XMLSchemaException("rcase-RecurseUnordered.1", new Object[] { Integer.toString(min1), (max1 == -1) ? "unbounded" : Integer.toString(max1), Integer.toString(min2), (max2 == -1) ? "unbounded" : Integer.toString(max2) });
        }
        int count1 = dChildren.size();
        int count2 = bChildren.size();
        boolean[] foundIt = new boolean[count2];
        int i = 0;
    Label_0103:
        while (i < count1) {
            XSParticleDecl particle1 = (XSParticleDecl)dChildren.get(i);
            int j = 0;
            while (j < count2) {
                XSParticleDecl particle2 = (XSParticleDecl)bChildren.get(j);
                try {
                    particleValidRestriction(particle1, dSGHandler, particle2, bSGHandler);
                    if (foundIt[j]) {
                        throw new XMLSchemaException("rcase-RecurseUnordered.2", null);
                    }
                    foundIt[j] = true;
                }
                catch (XMLSchemaException ex) {
                    ++j;
                    continue;
                }
                ++i;
                continue Label_0103;
            }
            throw new XMLSchemaException("rcase-RecurseUnordered.2", null);
        }
        int k = 0;
        while (k < count2) {
            XSParticleDecl particle3 = (XSParticleDecl)bChildren.get(k);
            if (!foundIt[k] && !particle3.emptiable()) {
                throw new XMLSchemaException("rcase-RecurseUnordered.2", null);
            }
            ++k;
        }
        return;
    }
    
    private static void checkRecurseLax(List<XSParticleDecl> dChildren, int min1, int max1, SubstitutionGroupHandler dSGHandler, List<XSParticleDecl> bChildren, int min2, int max2, SubstitutionGroupHandler bSGHandler) throws XMLSchemaException {
        if (!checkOccurrenceRange(min1, max1, min2, max2)) {
            throw new XMLSchemaException("rcase-RecurseLax.1", new Object[] { Integer.toString(min1), (max1 == -1) ? "unbounded" : Integer.toString(max1), Integer.toString(min2), (max2 == -1) ? "unbounded" : Integer.toString(max2) });
        }
        int count1 = dChildren.size();
        int count2 = bChildren.size();
        int current = 0;
        int i = 0;
    Label_0100:
        while (i < count1) {
            XSParticleDecl particle1 = (XSParticleDecl)dChildren.get(i);
            int j = current;
            while (j < count2) {
                XSParticleDecl particle2 = (XSParticleDecl)bChildren.get(j);
                ++current;
                try {
                    if (particleValidRestriction(particle1, dSGHandler, particle2, bSGHandler)) {
                        --current;
                    }
                }
                catch (XMLSchemaException ex) {
                    ++j;
                    continue;
                }
                ++i;
                continue Label_0100;
            }
            throw new XMLSchemaException("rcase-RecurseLax.2", null);
        }
        return;
    }
    
    private static void checkMapAndSum(List<XSParticleDecl> dChildren, int min1, int max1, SubstitutionGroupHandler dSGHandler, List<XSParticleDecl> bChildren, int min2, int max2, SubstitutionGroupHandler bSGHandler) throws XMLSchemaException {
        if (!checkOccurrenceRange(min1, max1, min2, max2)) {
            throw new XMLSchemaException("rcase-MapAndSum.2", new Object[] { Integer.toString(min1), (max1 == -1) ? "unbounded" : Integer.toString(max1), Integer.toString(min2), (max2 == -1) ? "unbounded" : Integer.toString(max2) });
        }
        int count1 = dChildren.size();
        int count2 = bChildren.size();
        int i = 0;
    Label_0097:
        while (i < count1) {
            XSParticleDecl particle1 = (XSParticleDecl)dChildren.get(i);
            int j = 0;
            while (j < count2) {
                XSParticleDecl particle2 = (XSParticleDecl)bChildren.get(j);
                try {
                    particleValidRestriction(particle1, dSGHandler, particle2, bSGHandler);
                }
                catch (XMLSchemaException ex) {
                    ++j;
                    continue;
                }
                ++i;
                continue Label_0097;
            }
            throw new XMLSchemaException("rcase-MapAndSum.1", null);
        }
        return;
    }
    
    public static boolean overlapUPA(XSElementDecl element1, XSElementDecl element2, SubstitutionGroupHandler sgHandler) {
        if (element1.fName == element2.fName && element1.fTargetNamespace == element2.fTargetNamespace) {
            return true;
        }
        XSElementDecl[] subGroup = sgHandler.getSubstitutionGroup(element1);
        int i = subGroup.length - 1;
        while (i >= 0) {
            if (subGroup[i].fName == element2.fName && subGroup[i].fTargetNamespace == element2.fTargetNamespace) {
                return true;
            }
            --i;
        }
        subGroup = sgHandler.getSubstitutionGroup(element2);
        i = subGroup.length - 1;
        while (i >= 0) {
            if (subGroup[i].fName == element1.fName && subGroup[i].fTargetNamespace == element1.fTargetNamespace) {
                return true;
            }
            --i;
        }
        return false;
    }
    
    public static boolean overlapUPA(XSElementDecl element, XSWildcardDecl wildcard, SubstitutionGroupHandler sgHandler) {
        if (wildcard.allowNamespace(element.fTargetNamespace)) {
            return true;
        }
        XSElementDecl[] subGroup = sgHandler.getSubstitutionGroup(element);
        int i = subGroup.length - 1;
        while (i >= 0) {
            if (wildcard.allowNamespace(subGroup[i].fTargetNamespace)) {
                return true;
            }
            --i;
        }
        return false;
    }
    
    public static boolean overlapUPA(XSWildcardDecl wildcard1, XSWildcardDecl wildcard2) {
        XSWildcardDecl intersect = wildcard1.performIntersectionWith(wildcard2, wildcard1.fProcessContents);
        return intersect == null || intersect.fType != 3 || intersect.fNamespaceList.length != 0;
    }
    
    public static boolean overlapUPA(Object decl1, Object decl2, SubstitutionGroupHandler sgHandler) {
        if ((decl1 instanceof XSElementDecl)) {
            if ((decl2 instanceof XSElementDecl)) {
                return overlapUPA((XSElementDecl)decl1, (XSElementDecl)decl2, sgHandler);
            }
            return overlapUPA((XSElementDecl)decl1, (XSWildcardDecl)decl2, sgHandler);
        }
        else {
            if ((decl2 instanceof XSElementDecl)) {
                return overlapUPA((XSElementDecl)decl2, (XSWildcardDecl)decl1, sgHandler);
            }
            return overlapUPA((XSWildcardDecl)decl1, (XSWildcardDecl)decl2);
        }
    }
    
    static {
        STRING_TYPE = (XSSimpleType)SchemaGrammar.SG_SchemaNS.getGlobalTypeDecl("string");
        XSConstraints.fEmptyParticle = null;
        ELEMENT_PARTICLE_COMPARATOR = new Comparator<XSParticleDecl>() {
            @Override
            public int compare(XSParticleDecl o1, XSParticleDecl o2) {
                XSParticleDecl pDecl1 = o1;
                XSParticleDecl pDecl2 = o2;
                XSElementDecl decl1 = (XSElementDecl)pDecl1.fValue;
                XSElementDecl decl2 = (XSElementDecl)pDecl2.fValue;
                String namespace1 = decl1.getNamespace();
                String namespace2 = decl2.getNamespace();
                String name1 = decl1.getName();
                String name2 = decl2.getName();
                boolean sameNamespace = namespace1 == namespace2;
                int namespaceComparison = 0;
                if (!sameNamespace) {
                    if (namespace1 != null) {
                        if (namespace2 != null) {
                            namespaceComparison = namespace1.compareTo(namespace2);
                        }
                        else {
                            namespaceComparison = 1;
                        }
                    }
                    else {
                        namespaceComparison = -1;
                    }
                }
                return (namespaceComparison != 0) ? namespaceComparison : name1.compareTo(name2);
            }
        };
    }
}
